/// @file toon-example.cpp
/// @brief Example program for QVision library. Applies corner detection over an input video.
///
/// This program will receive a path or url to a video file, and then will show hot points or corners of each frame in a
/// window. It detects corners where the minimal eigenvalue of the first derivatives matrix for a pixel is greater than a
/// Threshold value.
/// To use the program, simply compile, and execute the binary:
///
///	./qcorners --Video=<path or url to video>
///
/// @author PARP Group (http://perception.inf.um.es), Univ. of Murcia (Spain)
///
/// TooN webpage: http://mi.eng.cam.ac.uk/~twd20/Numerics/index.html
/// Download latest version from
///
///	http://mi.eng.cam.ac.uk/~er258/cvd/
///
/// about the bottom of the page, and un-tgz in this directory. Documentation is at
///
///	http://mi.eng.cam.ac.uk/~er258/cvd/toon/html/
///
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvgui/qvisioninterface.h>
#include <qvmplayercamera/qvmplayercamera.h>
#include <qvcore/qvapplication.h>

#include <TooN/numerics.h>	// Basic TooN library
#include <TooN/numhelpers.h> // If you want a few extras
#include <TooN/SVD.h>        // If you want singular value decompositions
#include <TooN/LU.h>         // If you want lu triangular decompositions
#include <TooN/SymEigen.h>   // If you want Eigen decomposition of symmetric matrices

class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name);
		void worker();
	};

MyWorker::MyWorker(QString name): QVWorker(name)
	{
	}

void MyWorker::worker()
	{
	qDebug() << "MyWorker::worker()";

	Vector<3> v1((double[]){1,2,3});
	Matrix<3,4> M((double[]){1,2,3,4,5,6,7,8,9,10,11,12});
	Vector<4> v2((double[]){1,2,3,4});

	std::cout << "v1 = " << v1 << std::endl;
	std::cout << "M = " << std::endl << M ;
	std::cout << "v2 = " << v2 << std::endl;

	std::cout << "v1 * M * v2 = " << v1*M*v2 << std::endl;

	qDebug() << "MyWorker::worker(): <- return";
	}

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Applies corner detection over an input video."

		);

	MyWorker worker("Corners Worker");

	QVisionInterface interface;

	int status = app.exec();

	qDebug() << "QVApplication.exec() returned" << status;

	return status;
	}
